list(
  # Adjustment name ----
  Name = "Multiple Imputations",

  # Input parameters to the adjustment function ----
  Parameters = list(
    # Parameter 1: a specification with label, type and default value
    nimp = list(
      label = "Number of imputations",
      value = 2L,
      input = "numeric"),
    # Parameter 2
    nburn = list(
      label = "Number of burn-in iterations",
      value = 100L,
      input = "numeric"),
    # Parameter 3
    nbetween = list(
      label = "Number of iterations between two successive imputations",
      value = 100L,
      input = "numeric"),
    # Parameter 4
    nsdf = list(
      label = "Number of degrees of freedom for spline",
      value = 6L,
      input = "numeric"),
    # Parameter 5
    runInParallel = list(
      label = "Run in parallel",
      value = FALSE,
      input = "checkbox")
  ),

  # Names of packages that must be made available to the adjustment function ----
  RequiredPackageNames = c("jomo"),

  ## Adjustment function ----
  AdjustmentFunction = function(inputData, parameters) {

    # Perform imputations per data set.
    # This is the actual worker function.
    workerFunction <- function(dataSet, nburn, nbetween, nimp, nsdf) {

      # Define covariates
      xColNames <- c("AIDS")
      # Define outcomes
      yColNames <- c("Age", "SqCD4", "Transmission", "Migr")

      # Create splines with proper names and intercept
      splineBasisMatrix <- as.data.table(splines::ns(dataSet$DY, df = nsdf))
      setnames(splineBasisMatrix, paste0("V", colnames(splineBasisMatrix)))
      intercept <- 1L

      # Define covariates of joint imputation model
      X <- cbind(dataSet[, ..xColNames],
                 Intercept = intercept,
                 splineBasisMatrix)
      # Define outcomes of joint imputation model
      Y <- dataSet[, ..yColNames]

      # Workaround for jomo bug:
      # Not used levels must be removed
      X <- droplevels(X)
      Y <- droplevels(Y)

      # Run model
      imp <- setDT(jomo::jomo(Y = Y,
                              X = X,
                              nburn = nburn,
                              nbetween = nbetween,
                              nimp = nimp,
                              out.iter = 100))

      indexColNames <- c("Imputation", "id")
      impColNames <- union(indexColNames, yColNames)
      dataSetColNames <- setdiff(colnames(dataSet), yColNames)

      mi <- cbind(imp[, ..impColNames], dataSet[, ..dataSetColNames])

      setcolorder(mi, union(indexColNames, names(dataSet)))

      return(mi)
    }

    # 1. Save original order for later
    inputData[, OrigSort := .I]

    # 2. Measures years from earlier diagnosis year
    inputData[, DY := DateOfDiagnosisYear - min(DateOfDiagnosisYear)]

    # 3. Split by gender to data sets
    dataSets <- inputData[Gender != ""]
    dataSets <- split(dataSets, by = c("Gender"))

    # 4. Execute the worker function per data set
    if (parameters$runInParallel) {
      # Run in parallel
      cl <- parallel::makeCluster(2)
      parallel::clusterExport(cl, "libPaths")
      parallel::clusterEvalQ(cl, {
        # .libPaths("D:/_REPOSITORIES/hiv-estimates-accuracy/packrat/lib/x86_64-w64-mingw32/3.4.2")
        .libPaths(libPaths)
        library(data.table)
      })
      outputData <- parallel::parLapply(cl,
                                        dataSets,
                                        workerFunction,
                                        nburn = parameters$nburn,
                                        nbetween = parameters$nbetween,
                                        nimp = parameters$nimp,
                                        nsdf = parameters$nsdf)
      parallel::stopCluster(cl)
    } else {
      # Run sequentially
      outputData <- lapply(dataSets,
                           workerFunction,
                           nburn = parameters$nburn,
                           nbetween = parameters$nbetween,
                           nimp = parameters$nimp,
                           nsdf = parameters$nsdf)
    }

    # 5. Combine all data sets
    outputData <- rbindlist(outputData)

    # 6. Restore original order per Imputation
    setorder(outputData, Imputation, OrigSort)

    return(outputData)
  }
)
